# -*- coding: utf-8 -*-
#
# File: PlominoReplicationManager.py
#
# Copyright (c) 2008 by ['Eric BREHAULT']
#
# Zope Public License (ZPL)
#

__author__ = """Xavier PERROT  - Eric BREHAULT <eric.brehault@makina-corpus.org>"""
__docformat__ = 'plaintext'

from Products.CMFPlomino.config import *

from DateTime import DateTime
from cStringIO import StringIO
from Acquisition import *
from index.PlominoIndex import PlominoIndex
from HttpUtils import authenticateAndLoadURL, authenticateAndPostToURL
from Products.CMFCore.utils import getToolByName
from Products.CMFPlomino.exceptions import PlominoReplicationException
from Products.CMFPlomino.PlominoUtils import StringToDate
import re
from Persistence import Persistent
from xml.dom.minidom import getDOMImplementation
from xml.dom.minidom import parseString
import xmlrpclib
import codecs
import os
import sys
import glob
import transaction

from ZPublisher.HTTPRequest import FileUpload
import csv

import logging
logger=logging.getLogger("Replication")

REMOTE_DOC_ID_SEPARATOR = '#'
REMOTE_DOC_DATE_SEPARATOR = '@'
REMOTE_DOC_IDS_HEADER = 'REMOTE_DOC_IDS'
REMOTE_URL_ADDED = 'url to replicate with'
REPLICATION_TYPES = {'push' : 'push', 'pull' : 'pull', 'pushpull' : 'push and pull'}
CONFLICT_RESOLUTION_TYPE = {'localwins' : 'local wins', 'remotewins' : 'remote wins', 'lastwins' : 'last wins'}
REPLICATION_MODES = {'view' : 'view', 'edit' : 'edit', 'add' : 'add'}
PASSWORD_DISPLAY_CAR = '*'
PLOMINO_IMPORT_SEPARATORS = {'semicolon (;)' : ';',
                             'comma (,)' : ',', 
                             'tabulation' : '\t', 
                             'white space' : ' ',
                             'end of line' : '\n',
                             'dash (-)' : '-'}

class PlominoReplicationManager(Persistent):
    """Plomino replication push/pull features
    """
    security = ClassSecurityInfo()

    # Methods
    security.declareProtected(EDIT_PERMISSION, 'manage_replications')
    def manage_replications(self, REQUEST=None):
        """replication form manager
        """
        #init
        infoMsg = ''
        error = False
        actionType = REQUEST.get('actionType', None)

        if actionType=='add':

            if not self.getReplicationEditingId():
                #new replication
                replication = self.newReplication()
                replication['mode'] = 'add'
                self.setReplication(replication)

        elif actionType=='cancel':

            #get the current replication editing
            replicationEditingId = self.getReplicationEditingId()
            #get replication
            replication = self.getReplication(replicationEditingId)
            if replication:
                if replication['mode'] == 'add':
                    self.deleteReplications([replicationEditingId])
                else:
                    replication['mode'] == 'view'
                    self.setReplication(replication)

        elif actionType=='save':

            #save params
            try:
                infoMsg = infoMsg + self.saveReplication(REQUEST) + MSG_SEPARATOR
            except PlominoReplicationException, e:
                infoMsg = infoMsg + 'error while saving : %s' % (e) + MSG_SEPARATOR 
                error = True    
        else:
            #actions on selection
            replicIds = REQUEST.get('selection', None)
            if replicIds:
                #check if simple url
                if type(replicIds)==str:
                    replicIds=[replicIds]

                #get the replications
                replications = self.getReplications()

                if actionType=='replicate':
                    #launch replications
                    for replicId in replicIds:

                        error = False
                        try:
                            infoMsg = self.replicate(replicId)
                        except PlominoReplicationException, e:
                            infoMsg = 'error while replicating ' + replicId + ' : %s' % (e) + MSG_SEPARATOR
                            error = True

                        #write message
                        self.writeMessageOnPage(infoMsg, REQUEST, error)

                    #no end message
                    error = False
                    infoMsg = ''

                elif actionType=='delete':
                    try:
                        infoMsg = self.deleteReplications(replicIds)
                    except PlominoReplicationException, e:
                        infoMsg = infoMsg + 'error while deleting all : %s' % (e) + MSG_SEPARATOR
                        error = True

                elif actionType=='edit':

                    #set mode to edit
                    if not self.getReplicationEditingId():
                        replicId = replicIds[0]
                        replication = self.getReplication(replicId)
                        replication['mode'] = 'edit'
                        self.setReplication(replication)

                else:
                    infoMsg = infoMsg + actionType + ' : unmanaged action' + MSG_SEPARATOR
                    error = True
            else:
                infoMsg = infoMsg + 'empty selection' + MSG_SEPARATOR
                error = True

        #write message
        self.writeMessageOnPage(infoMsg, REQUEST, error)

        #redirect
        REQUEST.RESPONSE.redirect(self.absolute_url()+'/DatabaseReplication')

    security.declareProtected(EDIT_PERMISSION, 'deleteReplications')
    def deleteReplications(self, replicationIds):
        """delete remoteUrl list        
        """
        #init
        infoMsg = ''
        error = False

        #replications
        replications = self.getReplications()

        for replicationId in replicationIds:
            try:
                replications.pop(replicationId)
                infoMsg = infoMsg + 'replication ' + str(replicationId) + ' deleted' + MSG_SEPARATOR
            except PlominoReplicationException, e:
                infoMsg = infoMsg + 'error while deletion of replication ' + replicationId + ' : %s' % (e) + MSG_SEPARATOR
                error = True

        if error:
            raise PlominoReplicationException, infoMsg

        #save replications
        self.setReplications(replications)

        return infoMsg

    security.declareProtected(EDIT_PERMISSION, 'saveReplication')
    def saveReplication(self, REQUEST=None):
        """save replication        
        """
        try:
            replication = self.buidReplicationFromRequest(REQUEST)
        except PlominoReplicationException, e:
            infoMsg = 'error while checking parameters : %s' % (e) + MSG_SEPARATOR
            raise PlominoReplicationException, infoMsg 

        #mode
        replication['mode'] = 'view'

        #save
        self.setReplication(replication)

        #result
        return "replication saved"

    security.declareProtected(EDIT_PERMISSION, 'replicate')
    def replicate(self, replicationId=None):
        """launch replication with just remote url passed
        """
        #check param
        if not replicationId:
            raise PlominoReplicationException, 'Replication id required'

        #get replication
        replication = self.getReplication(replicationId)
        if not replication:
            raise PlominoReplicationException, 'Unknown replication id'

        try:
            infoMsg = replication['remoteUrl'] + ' : ' + self.launchReplication(replication)
        except PlominoReplicationException, e:
            infoMsg = 'error while replicating ' + replication['remoteUrl'] + ' : %s' % (e) + MSG_SEPARATOR
            error = True

        #launch replication
        return infoMsg        

    security.declareProtected(EDIT_PERMISSION, 'launchReplication')
    def launchReplication(self, replication):
        """launch replication with params
        """

        #init
        infoMsg = ''
        error = False
        nbDocPushed = 0
        nbDocNotPushed = 0
        nbDocPulled = 0
        nbDocNotPulled = 0
        lastReplicationDatePush = None
        lastReplicationDatePull = None

        #check replication
        try:
            self.checkReplication(replication)
        except PlominoReplicationException, e:
            infoMsg = infoMsg + 'error while cheking parameters : %s' % (e) + MSG_SEPARATOR
            error = True

        #get remote documents ->  id:lastEditDate
        if not error:
            try:
                remoteDocuments = self.getRemoteDocuments(replication)
            except PlominoReplicationException, e:
                infoMsg = infoMsg + 'error while getting remote documents : %s' % (e) + MSG_SEPARATOR
                error = True

        #get local documents
        if not error:
            if replication.has_key('restricttoview'):
                restricttoview=replication['restricttoview']
                if restricttoview is not None and not(restricttoview==''):
                    try:
                        localDocuments = self.getView(restricttoview).getAllDocuments()
                    except PlominoReplicationException, e:
                        infoMsg = infoMsg + 'error while getting local documents : %s' % (e) + MSG_SEPARATOR
                        error = True
                else:
                    localDocuments = self.getAllDocuments()
            else:
                localDocuments = self.getAllDocuments()

        #flag replication begin on remote
        if not error:
            try:
                authenticateAndLoadURL(replication['remoteUrl'] + '/startReplicationRemote?RemoteUrl='+ self.absolute_url() +'&repType='+replication['repType'],replication['username'],replication['password'])
            except Exception, e:
                infoMsg = infoMsg + 'error while flagging replication on remote : %s' % (e) + MSG_SEPARATOR
                error = True

        #flag replication begin on local
        lastReplicationDatePull = self.getReplicationDate(replication['remoteUrl'],'pull')
        lastReplicationDatePush = self.getReplicationDate(replication['remoteUrl'],'push')
        self.startReplication(replication['remoteUrl'],replication['repType'])   


        #push documents
        if not error:
            if replication['repType']=='push' or replication['repType']=='pushpull':
                for doc in localDocuments:
                    #check if document can be exported
                    if self.exportableDoc(doc, remoteDocuments, lastReplicationDatePush, replication['whoWins']):
                        #export
                        try:
                            self.exportDocumentPush(doc, replication['remoteUrl'], replication['username'], replication['password'])
                            #infoMsg = infoMsg + 'document ' + doc.getId() + ' : push done' + MSG_SEPARATOR
                            nbDocPushed = nbDocPushed + 1
                        except PlominoReplicationException, e:
                            infoMsg = infoMsg + doc.getId() + ' push error : %s' % (e) + MSG_SEPARATOR
                            error = True
                    else:
                        #infoMsg = infoMsg + 'document ' + doc.getId() + ' : not pushed (not an error)' + MSG_SEPARATOR
                        nbDocNotPushed = nbDocNotPushed + 1

        #pull documents
        if not error:
            if replication['repType']=='pull' or replication['repType']=='pushpull':
                for docId in remoteDocuments:
                    #check if document can be imported
                    if self.importableDoc(docId, remoteDocuments[docId], lastReplicationDatePull, replication['whoWins']):
                        #import
                        try:
                            self.importDocumentPull(docId, replication['remoteUrl'], replication['username'], replication['password'])
                            #infoMsg = infoMsg + 'document ' + docId + ' : pull done' + MSG_SEPARATOR
                            nbDocPulled = nbDocPulled + 1
                        except PlominoReplicationException, e:                            
                            infoMsg = infoMsg + docId + ' pull error : %s' % (e) + MSG_SEPARATOR
                            error = True
                    else:
                        #infoMsg = infoMsg + 'document ' + docId + ' : not pulled (not an error)' + MSG_SEPARATOR
                        nbDocNotPulled = nbDocNotPulled + 1

        #add doc pushed number
        if nbDocPushed == 0:
            infoMsg = infoMsg + 'no document pushed '
        elif nbDocPushed == 1:
            infoMsg = infoMsg + '1 document pushed '
        else:
            infoMsg = infoMsg + str(nbDocPushed) + ' documents pushed '
        infoMsg = infoMsg + '(not pushed : ' + str(nbDocNotPushed) + '), ' 

        #add doc pulled number
        if nbDocPulled == 0:
            infoMsg = infoMsg + 'no document pulled '
        elif nbDocPulled == 1:
            infoMsg = infoMsg + '1 document pulled '
        else:
            infoMsg = infoMsg + str(nbDocPulled) + ' documents pulled '
        infoMsg = infoMsg + '(not pulled : ' + str(nbDocNotPulled) + ') '

        #errors
        if error:
            raise PlominoReplicationException, infoMsg

        return infoMsg

    security.declareProtected(EDIT_PERMISSION, 'getRemoteDocuments')
    def getRemoteDocuments(self, replication):
        """return the database documents
        """
        self.checkReplication(replication)

        #get remote documents
        url=replication['remoteUrl'] + '/getDocumentsIds'
        if replication.has_key('restricttoview'):
            restricttoview=replication['restricttoview']
            if restricttoview is not None and not(restricttoview==''):
                url=url+"?restricttoview="+restricttoview
        remoteDocumentsIds = authenticateAndLoadURL(url, replication['username'], replication['password']).read()

        #check if starts with REMOTE_DOC_IDS_HEADER
        if not remoteDocumentsIds.startswith(REMOTE_DOC_IDS_HEADER):
            raise PlominoReplicationException, "Connection error"

        #string ids
        docs = remoteDocumentsIds.split(REMOTE_DOC_ID_SEPARATOR)
        #remove header
        docs.pop(0)
        docs.pop()
        result={}
        for d in docs:
            (docid, modifdate) = d.split(REMOTE_DOC_DATE_SEPARATOR)
            result[docid]=DateTime(modifdate)
        return result

    security.declareProtected(READ_PERMISSION, 'getDocumentsIds')
    def getDocumentsIds(self, REQUEST=None):
        """return the database documents ids in a string
        """
        if REQUEST is not None:
            restricttoview=REQUEST.get('restricttoview',None)
            if restricttoview is not None:
                docs = self.getView(restricttoview).getAllDocuments()
            else:
                docs = self.getAllDocuments()
        else:
            docs = self.getAllDocuments()
        ids = REMOTE_DOC_IDS_HEADER+REMOTE_DOC_ID_SEPARATOR
        for d in docs:
            ids=ids+d.id+REMOTE_DOC_DATE_SEPARATOR+d.getLastModified(asString=True)+REMOTE_DOC_ID_SEPARATOR
        return ids

    security.declarePrivate('exportableDoc')
    def exportableDoc(self, doc, remoteDocuments, lastReplicationDate, whowins):
        """check if document can be exported to remoteUrl
        """
        #initialization
        res = False
        lastEditRemoteDocumentDate = None

        #search for doc in remoteDocuments
        if doc.id in remoteDocuments:
            lastEditRemoteDocumentDate = remoteDocuments[doc.id]

        #no remoteDoc found -> export 
        if not lastEditRemoteDocumentDate:
            res = True
        else:
            #zope modification time (plone modification time is not 
            #set while document modified thru script)
            #TODO : UTC
            lastEditDocumentDate = doc.getLastModified()
            #check dates
            if not lastReplicationDate:
                #no replication before
                res = (whowins == 'localwins') or ((whowins == 'lastwins') and (lastEditRemoteDocumentDate < lastEditDocumentDate))
            elif (lastEditDocumentDate > lastReplicationDate):
                #check conflict
                if (lastEditRemoteDocumentDate > lastReplicationDate):
                    res = (whowins == 'localwins') or ((whowins == 'lastwins') and (lastEditRemoteDocumentDate < lastEditDocumentDate))
                else:
                   res = True

        #result
        return res

    security.declarePrivate('exportDocumentPush')
    def exportDocumentPush(self, doc, remoteUrl, username, password):
        """ exports document to remoteUrl
            send object as a xml stream via HTTP multipart POST
        """
        id=doc.id
        xmlstring=self.exportAsXML(docids=[id])
        result = authenticateAndPostToURL(remoteUrl+"/importFromXML", username, password, '%s.%s' % (id, 'xml'), xmlstring.encode('utf-8'))

    security.declarePrivate('importableDoc')
    def importableDoc(self, docId, lastEditRemoteDocumentDate, lastReplicationDate, whowins):
        """check if document can be imported to remoteUrl
        """
        res = False

        #search for doc locally
        localDoc = self.getDocument(docId)
                
        #if no local doc found -> import 
        if localDoc is None:
            res = True 
        else:
            #zope modification time (plone modification time is not 
            #set while document modified thru script)
            #TODO : UTC
            lastEditDocumentDate = localDoc.getLastModified()

            #check dates
            if not lastReplicationDate:
                #no replication before
                res = (whowins == 'remotewins') or ((whowins == 'lastwins') and (lastEditRemoteDocumentDate > lastEditDocumentDate))
            elif (lastEditRemoteDocumentDate > lastReplicationDate):
                #check conflict
                if (lastEditDocumentDate > lastReplicationDate):
                    res = (whowins == 'remotewins') or ((whowins == 'lastwins') and (lastEditRemoteDocumentDate > lastEditDocumentDate))
                else:
                   res = True

        #result
        return res

    security.declarePrivate('importDocumentPull')
    def importDocumentPull(self, id, remoteUrl, username, password):
        """ imports document from remoteurl
            send object as a .zexp stream via HTTP multipart POST
        """
        f=authenticateAndLoadURL(remoteUrl+"/exportAsXML?docids="+id, username, password)
        self.importFromXML(xmlstring=f.read())

    security.declareProtected(EDIT_PERMISSION, 'getReplications')
    def getReplications(self):
        """returns a hash map representing replication history
        key : id
        value : replication hash map 
        """
        if not (hasattr(self,'replicationHistory')):
            self.replicationHistory = {}
        return self.replicationHistory

    security.declareProtected(EDIT_PERMISSION, 'setReplications')
    def setReplications(self, replications):
        """sets the replications hashmap 
        """
        self.replicationHistory = replications

        self.managePlominoCronTab()

        return self.replicationHistory

    security.declareProtected(EDIT_PERMISSION, 'getReplicationEditingId')
    def getReplicationEditingId(self):
        """returns the replication Id being edited
        """
        res = None
        replications = self.getReplications()
        for replicationId in replications:
            replication = self.getReplication(replicationId)
            if replication['mode'] == 'edit' or replication['mode'] == 'add':
                res = replicationId            
        return res

    security.declareProtected(EDIT_PERMISSION, 'getReplication')
    def getReplication(self, id):
        """returns the replication Id being edited
        """
        res = None
        #replication list
        replications = self.getReplications()
        #serach id
        searchId = str(id)
        if replications.has_key(searchId):
            res = replications[searchId]
        return res


    security.declarePrivate('setReplication')
    def setReplication(self, replication):
        """add the replication to the the replication hash map
        """
        #check replication
        replication = self.checkReplication(replication)
                        
        #replication list
        replications = self.getReplications()        

        replications[str(replication['id'])] = replication

        #set
        self.setReplications(replications)

    security.declareProtected(EDIT_PERMISSION, 'getReplicationsDates')
    def getReplicationsDates(self):
        """returns a hash map representing replication history
        key : url
        value : hash map {push, pull} 
        """
        if not (hasattr(self,'replicationsDates')):
            self.replicationsDates = {}
        return self.replicationsDates

    security.declareProtected(EDIT_PERMISSION, 'setReplicationsDates')
    def setReplicationsDates(self, repDates):
        """sets the replications hashmap 
        """
        self.replicationsDates = repDates
        return self.replicationsDates

    security.declarePrivate('getReplicationDate')
    def getReplicationDate(self, remoteUrl, replicationtype):
        """returns the replication date for id and type
        """
        #test params
        if not replicationtype in REPLICATION_TYPES.keys():
            raise PlominoReplicationException, 'Unknown replication type "' + replication['repType'] + '" (' + str(REPLICATION_TYPES.keys()) + ' expected)'

        #push pull not allowed
        if replicationtype == 'pushpull':
            raise PlominoReplicationException, 'Unable to return two dates for push pull. Choose push or pull.'

        #get hash map
        repDates = self.getReplicationsDates()

        #test if remoteurl is in
        if not repDates.has_key(remoteUrl):
            return None
        elif not repDates[remoteUrl].has_key(replicationtype):
            return None
        else:
            #return date
            return repDates[remoteUrl][replicationtype]

    security.declarePrivate('setReplicationDate')
    def setReplicationDate(self, remoteUrl, replicationtype, date):
        """sets the replication date for url and type
        """
        if not replicationtype in REPLICATION_TYPES.keys():
            raise PlominoReplicationException, 'Unknown replication type "' + replication['repType'] + '" (' + str(REPLICATION_TYPES.keys()) + ' expected)'

        #get hash map
        repDates = self.getReplicationsDates()

        #get url dates
        if repDates.has_key(remoteUrl):
            repDate = repDates[remoteUrl]
        else:
            repDate = {'push' : None, 'pull' : None}

        #add parameters
        if replicationtype == 'pushpull':
            repDate['push'] = date
            repDate['pull'] = date
        else:
            repDate[replicationtype] = date

        #add it
        repDates[remoteUrl] = repDate

        #set
        self.setReplicationsDates(repDates)

    security.declarePublic('displayDates')
    def displayDates(self, remoteUrl, context):
        """returns a string for displaying dates on template
        """
        #init
        res = ''

        #date formater
        util = getToolByName(context, 'translation_service')

        #dates
        datePush = self.getReplicationDate(remoteUrl, 'push')
        datePull = self.getReplicationDate(remoteUrl, 'pull')
        if datePush:
            datePush = util.ulocalized_time(datePush, True, context, domain='plonelocales')
            res = 'last push : ' + str(datePush) + ', '
        else: 
            res = 'no last push, '
        if datePull:
            datePull = util.ulocalized_time(datePull, True, context, domain='plonelocales')
            res = res + 'last pull : ' + str(datePull)
        else: 
            res = res + 'no last pull'    
        return res

    security.declarePrivate('setReplicationMode')
    def setReplicationMode(self, remoteId, mode):
        """sets the replication mode for url 
        """
        #test params
        if (mode != 'view') and (mode != 'edit') and (mode != 'add'):
            raise PlominoReplicationException, "Replication mode expected : view, edit or add"

        replicationEditingId = self.getReplicationEditingId()

        if (replicationEditingId) and (mode =='edit'):
            raise PlominoReplicationException, "Multiple replication editing forbidden"

        if (replicationEditingId) and (mode =='add'):
            raise PlominoReplicationException, "Unable to add, another replication already editing or adding"

        #current replication
        replication = self.getReplication(replicId)
        if not replication:
            raise PlominoReplicationException, "Replication unknown (%s)" % (str(replicId))

        #add parameter
        replication['mode'] = mode

        #set
        self.setReplication(replication)

    security.declarePublic('startReplicationRemote')
    def startReplicationRemote(self, REQUEST = None):
        """flags the begining of the transaction        
        """
        if not REQUEST is None:
            remoteUrl = REQUEST.get('RemoteUrl',None)
            repType = REQUEST.get('repType',None)
        else:
            remoteUrl = self.absolute_url()
            repType = 'push'

        self.startReplication(remoteUrl, repType)


    security.declarePublic('startReplication')
    def startReplication(self,remoteUrl, repType):
        """flags the beginning of the transaction        
        """
        now = DateTime().toZone('UTC')
        self.setReplicationDate(remoteUrl, repType, now)

    security.declarePublic('hideIt')
    def hideIt(self, it, hideCar = PASSWORD_DISPLAY_CAR):
        """return an hidden string to display
        """
        if it:
            count = len(it)
        else:
            count = 0
        return hideCar*count

    security.declarePublic('displayItNice')
    def displayItNice(self, it, hashMapName):
        """return a nice display title for type passed
        """
        res = None
        if hashMapName == 'REPLICATION_TYPES':
            hashmap = REPLICATION_TYPES
        elif hashMapName == 'CONFLICT_RESOLUTION_TYPE':
            hashmap = CONFLICT_RESOLUTION_TYPE
        elif hashMapName == 'REPLICATION_MODES':
            hashmap = REPLICATION_MODES
        else:
            hashmap = None
        if hashmap:
            if hashmap.has_key(it):
                res = hashmap[it]         
        return res

    security.declarePublic('resetReplications')
    def resetReplications(self):
        """reset the replication hashmap
        """
        self.replicationHistory = {}
        self.replicationsDates = {}

    security.declarePrivate('getNewId')
    def getNewId(self):
        """return a new id
        """
        res = 0
        replications = self.getReplications()
        for replicId in replications:
            if replicId > res:
                res = int(replicId)
        return res + 1

    security.declarePrivate('newReplication')
    def newReplication(self):
        """return an empty replication
        """
        return self.buildReplication(self.getNewId(),'', REMOTE_URL_ADDED, '', '', 'pull', 'localwins', False, '', '', 'view')

    security.declarePrivate('createReplication')
    def buidReplicationFromRequest(self, REQUEST = None):
        """returns a replication from request
        """
        if not REQUEST:
            raise PlominoReplicationException, "Request is required" 

        #params
        id = REQUEST.get('replicationId', None)
        name = REQUEST.get('name', None)
        remoteUrl = REQUEST.get('remoteUrl', None)
        username = REQUEST.get('username', None)
        password = REQUEST.get('password', None)
        repType = REQUEST.get('repType', None)
        whoWins = REQUEST.get('whoWins', None)
        scheduled = REQUEST.get('scheduled', False)
        restricttoview = REQUEST.get('restricttoview', None)
        cron = REQUEST.get('cron', None)
        mode = REQUEST.get('mode', None)

        return self.buildReplication(id, name, remoteUrl, username, password, repType, whoWins, scheduled, cron, restricttoview, mode)

    security.declarePrivate('createReplication')
    def buildReplication(self, id, name, remoteUrl, username, password, repType, whoWins, scheduled, cron, restricttoview, mode):
        """return an empty replication
        """
        return self.checkReplication({'id': id,
                                    'name' : name,
                                    'remoteUrl' : remoteUrl,
                                    'username' : username,
                                    'password' : password,
                                    'repType' : repType,
                                    'whoWins' : whoWins,
                                    'scheduled' : scheduled,
                                    'restricttoview':restricttoview,
                                    'cron':cron,
                                    'mode':mode})

    security.declarePrivate('checkReplication')
    def checkReplication(self, replication):
        """tests the replication params
        """
        try:
            if not replication:
                raise PlominoReplicationException, "replication empty"

            if not replication.has_key('id'):
                raise PlominoReplicationException, "id not set"
            if not replication.has_key('name'):
                raise PlominoReplicationException, "name not set"
            if not replication.has_key('remoteUrl'):
                raise PlominoReplicationException, "remote Url not set"
            if not replication.has_key('username'):
                raise PlominoReplicationException, "username not set"
            if not replication.has_key('password'):
                raise PlominoReplicationException, "password not set"
            if not replication.has_key('repType'):
                raise PlominoReplicationException, "replication type not set"
            if not replication.has_key('whoWins'):
                raise PlominoReplicationException, "conflict resolution type not set"
            if not replication.has_key('scheduled'):
                raise PlominoReplicationException, "scheduled type not set"
            if not replication.has_key('restricttoview'):
                raise PlominoReplicationException, "restricttoview not set"
            if not replication.has_key('cron'):
                raise PlominoReplicationException, "cron not set"
            if not replication.has_key('mode'):
                raise PlominoReplicationException, "mode not set"

            if replication['mode'] != 'add' and not replication['id']:
                raise PlominoReplicationException, "id required"

            #name not checked

            if not replication['remoteUrl']:
                raise PlominoReplicationException, 'Remote URL required'
            if replication['remoteUrl'] == self.absolute_url():
                raise PlominoReplicationException, 'Replication on current base forbidden'

            #user name and pwd not checked

            if not replication['repType'] in REPLICATION_TYPES.keys():
                raise PlominoReplicationException, 'Unknown replication type "' + replication['repType'] + '" (' + str(REPLICATION_TYPES.keys()) + ' expected)'

            if not replication['whoWins'] in CONFLICT_RESOLUTION_TYPE.keys():
                raise PlominoReplicationException, 'Unknown conflict resolution type "' + replication['whoWins'] + '" (' + str(CONFLICT_RESOLUTION_TYPE.keys()) + ' expected)'

            if replication['scheduled'] and (not replication['cron'] or len(replication['cron'])==0):
                raise PlominoReplicationException, 'Cron required if scheduled'

            if not replication['mode'] in REPLICATION_MODES.keys():
                raise PlominoReplicationException, 'Unknown replication mode "' + replication['mode'] + '" (' + str(REPLICATION_MODES.keys()) + ' expected)'

            return replication

        except PlominoReplicationException, e:
            raise PlominoReplicationException, 'replication has problem : %s' % (e)

    security.declareProtected(EDIT_PERMISSION, 'manage_importation')
    def manage_importation(self, REQUEST=None):
        """CSV import form manager
        """

        #init
        infoMsg = ''
        error = False
        actionType = REQUEST.get('actionType', None)

        #clear report
        self.resetReport()

        if actionType=='import':

            #get file name
            try:
                infoMsg = self.processImport(REQUEST)
            except PlominoReplicationException, e:
                infoMsg = infoMsg + 'error while importing : %s' % (e) + MSG_SEPARATOR 
                error = True

        elif actionType=='reset_report':
            self.resetReport()
        else:
            infoMsg = infoMsg + actionType + ' : unmanaged action' + MSG_SEPARATOR
            error = True

        #write message
        self.writeMessageOnPage(infoMsg, REQUEST, error)

        #redirect
        REQUEST.RESPONSE.redirect(self.absolute_url()+'/DatabaseReplication')

    security.declarePublic('getSeparators')
    def getSeparators(self):
        """returns the separator list
        """
        return PLOMINO_IMPORT_SEPARATORS

    security.declareProtected(EDIT_PERMISSION, 'processImport')
    def processImport(self, REQUEST):
        """
        Process the importation.
        """
        #result message
        infoMsg = ''

        if REQUEST:
            formName = REQUEST.get('formSelected',None)
            separatorName = REQUEST.get('separator',None)
            fileToImport = REQUEST.get('import_file',None)
            file_encoding = REQUEST.get('file_encoding', "utf-8")

        #form name
        if not formName:
            raise PlominoReplicationException, 'form required'

        #form obj
        formObj = self.getForm(formName)
        if not formObj:
            raise PlominoReplicationException, 'form ' + formName +  ' is unknown'

        #separator
        if not separatorName:
            raise PlominoReplicationException, 'separator required'
        separator = None
        if PLOMINO_IMPORT_SEPARATORS.has_key(separatorName):
            separator = PLOMINO_IMPORT_SEPARATORS[separatorName]

        #file
        if not fileToImport:
            raise PlominoReplicationException, 'file required'

        #check type
        if not isinstance(fileToImport, FileUpload):
            raise PlominoReplicationException, 'unrecognized file uploaded'

        #parse
        fileContent = self.parseFile(fileToImport, formName, separator, file_encoding)

        #import
        nbDocDone, nbDocFailed = self.importCsv(fileContent)

        #message
        infoMsg = fileToImport.filename + ' processed : ' + str(nbDocDone) + ' document(s) imported, ' + str(nbDocFailed) + ' document(s) failed.'
        return infoMsg

    security.declareProtected(EDIT_PERMISSION, 'processImport')
    def processImportAPI(self, formName, separator, fileToImport, file_encoding='utf-8'):
        """
        Process import API method.
        """

        #form name
        if not formName:
            raise PlominoReplicationException, 'form required'

        #form obj
        formObj = self.getForm(formName)
        if not formObj:
            raise PlominoReplicationException, 'form ' + formName +  ' is unknown'

        #separator
        if not separator:
            raise PlominoReplicationException, 'separator required'

        #file
        if not fileToImport:
            raise PlominoReplicationException, 'file required'

        #check type
        #if not isinstance(fileToImport, File):
         #   raise PlominoReplicationException, 'unrecognized file uploaded'

        #parse and import
        fileContent = self.parseFile(fileToImport, formName, separator, file_encoding)

        nbDocDone, nbDocFailed = self.importCsv(fileContent)        
        logger.info('Import processed : ' + str(nbDocDone) + ' document(s) imported, ' + str(nbDocFailed) + ' document(s) failed.')       

    security.declareProtected(EDIT_PERMISSION, 'parseFile')
    def parseFile(self, fileToImport, formName, separator, file_encoding='utf-8'):
        """
        """
        res= []
        try:
            if not fileToImport:
                raise PlominoReplicationException, 'file not set'

            if not separator:
                raise PlominoReplicationException, 'separator not set'

            #use the python cvs module
            reader = csv.DictReader(fileToImport.readlines(), delimiter=separator)

            #add the form name and copy reader values
            for line in reader:
                docInfos = {}

                #add form name
                docInfos['Form'] = formName

                #copy col values
                for col in line:
                    v = line[col]
                    if v is None:
                        v = u''
                    docInfos[col] = v.decode(file_encoding)

                #add doc infos to res
                res.append(docInfos)

            #result
            return res

        except PlominoReplicationException, e:
            raise PlominoReplicationException, 'error while parsing file (%s)' % (e)            

    security.declareProtected(EDIT_PERMISSION, 'importCsv')
    def importCsv(self, fileContent):
        """
        Import csv from content parsed.
        """

        #import
        nbDocDone = 0
        nbDocFailed = 0
        counter = 0

        i = 2
        logger.info("Documents count: %d" % len(fileContent))
        txn = transaction.get()
        for docInfos in fileContent:
            try:
                #create doc
                doc = self.createDocument()
                #fill it
                form = self.getForm(docInfos['Form'])
                form.readInputs(doc, docInfos, process_attachments=False, applyhidewhen=False)
                doc.setItem('Form', docInfos['Form'])
                #add items that don't correspond to any field 
                computedItems = doc.getItems()
                for info in docInfos:
                    if info not in computedItems:
                        v = docInfos[info]
                        if isinstance(v, str):
                            v.decode('utf-8')
                        doc.setItem(info, v)
                #save
                doc.save(creation=True, refresh_index=True)
                #count
                nbDocDone = nbDocDone + 1

            except PlominoReplicationException, e:
                nbDocFailed = nbDocFailed + 1
                self.addToReport(i, e, True)
            #next 
            i = i + 1
            counter = counter + 1
            if counter == 100:
                txn.savepoint(optimistic=True)
                counter = 0
                logger.info("%d documents imported successfully, %d errors(s) ...(still running)" % (nbDocDone, nbDocFailed))
        txn.commit()
        logger.info("Importation finished: %d documents imported successfully, %d document(s) not imported" % (nbDocDone, nbDocFailed)) 
        
        #result sent
        return (nbDocDone, nbDocFailed)

    security.declarePublic('getReport')
    def getReport(self):
        """returns last importation report
        """
        if not (hasattr(self,'importReport')):
            self.importReport = None
        return self.importReport

    security.declareProtected(EDIT_PERMISSION, 'resetReport')
    def resetReport(self):
        """clears report
        """
        self.importReport = None

    security.declareProtected(EDIT_PERMISSION, 'addToReport')
    def addToReport(self, lineNumber, infoMessage, error = False):
        """ adds entry to report
        """
        if not self.getReport():
            self.importReport = []
        if error:
            state = 'error'
        else:
            state = 'success'

        self.importReport.append({'line' : lineNumber, 'state' : state, 'infoMsg' : infoMessage})

    security.declareProtected(READ_PERMISSION, 'manage_exportAsXML')
    def manage_exportAsXML(self, REQUEST):
        """
        """
        restricttoview = REQUEST.get("restricttoview", "")
        if restricttoview != "":
            sourceview = self.getView(restricttoview)
            docids = [b.id for b in sourceview.getAllDocuments(getObject=False)]
        else:
            docids = None
        if REQUEST.get('targettype') == "file":
            REQUEST.RESPONSE.setHeader('content-type', 'text/xml')
            if restricttoview:
                label = restricttoview
            else:
                label = "all"
            REQUEST.RESPONSE.setHeader("Content-Disposition", "attachment; filename=%s-%s-documents.xml" % (self.id, label))
        return self.exportAsXML(docids, REQUEST=REQUEST)

    security.declareProtected(READ_PERMISSION, 'exportAsXML')
    def exportAsXML(self, docids=None, targettype='file', targetfolder='', REQUEST=None):
        """ Export documents to XML.
        The targettype can be file or folder.
        """
        if REQUEST:
            targettype=REQUEST.get('targettype', 'file')
            targetfolder=REQUEST.get('targetfolder')
            str_docids=REQUEST.get("docids")
            if str_docids is not None:
                docids = str_docids.split("@")

        impl = getDOMImplementation()

        if docids is None:
            docs = self.getAllDocuments()
        else:
            docs = [self.getDocument(id) for id in docids]

        if targettype == 'file':
            xmldoc = impl.createDocument(None, "plominodatabase", None)
            root = xmldoc.documentElement
            root.setAttribute("id", self.id)

            for d in docs:
                node = self.exportDocumentAsXML(xmldoc, d)
                root.appendChild(node)

            if REQUEST is not None:
                REQUEST.RESPONSE.setHeader('content-type', 'text/xml')
            return xmldoc.toxml()

        if targettype == 'folder':

            if REQUEST:
                targetfolder=REQUEST.get('targetfolder')

            exportpath = os.path.join(targetfolder,(self.id))
            if os.path.isdir(exportpath):
                # remove previous export
                for f in glob.glob(os.path.join(exportpath,"*.xml")):
                    os.remove(f)
            else:
                os.makedirs(exportpath)

            for d in docs:
                docfilepath = os.path.join(exportpath, (d.id+'.xml'))
                #DBG if os.path.exists(docfilepath):
                #DBG     logger.info("Skipping %s"%docfilepath)
                #DBG     continue
                logger.info("Exporting %s"%docfilepath)
                xmldoc = impl.createDocument(None, "plominodatabase", None)
                root = xmldoc.documentElement
                root.setAttribute("id", d.id)
                node = self.exportDocumentAsXML(xmldoc, d)
                root.appendChild(node)
                xmlstring = xmldoc.toxml()
                self.saveFile(docfilepath, xmlstring)

    @staticmethod
    def saveFile(path, content):
        fileobj = codecs.open(path, "w", "utf-8")
        fileobj.write(content)
        fileobj.close()

    security.declareProtected(READ_PERMISSION, 'exportDocumentAsXML')
    def exportDocumentAsXML(self, xmldoc, doc):
        """
        """
        node = xmldoc.createElement('document')
        node.setAttribute('id', doc.id)
        node.setAttribute('lastmodified', doc.getLastModified(asString=True))

        # export items
        str_items = xmlrpclib.dumps((doc.items,), allow_none=True)
        dom_items = parseString(str_items)
        node.appendChild(dom_items.documentElement)

        # export attached files
        for f in doc.getFilenames():
            attached_file = doc.getfile(f)
            fnode = xmldoc.createElement('attachment')
            fnode.setAttribute('id', f)
            fnode.setAttribute('contenttype', getattr(attached_file, 'content_type',''))
            data = xmldoc.createCDATASection(str(attached_file).encode('base64'))
            fnode.appendChild(data)
            node.appendChild(fnode)

        return node

    security.declareProtected(REMOVE_PERMISSION, 'manage_importFromXML')
    def manage_importFromXML(self, REQUEST):
        """
        """
        (imports, errors) = self.importFromXML(REQUEST=REQUEST)
        self.writeMessageOnPage("%d documents imported successfully, %d document(s) not imported" % (imports, errors), REQUEST, False)
        REQUEST.RESPONSE.redirect(self.absolute_url()+"/DatabaseReplication")

    security.declareProtected(REMOVE_PERMISSION, 'importFromXML')
    def importFromXML(self, xmlstring=None, sourcetype='sourceFile', file=None, from_folder=None, REQUEST=None):
        """ Import documents from XML.
        The sourcetype can be sourceFile or sourceFolder.
        """
        logger.info("Start documents import")
        self.setStatus("Importing documents (0%)")
        xml_files = []
        txn = transaction.get()
        if REQUEST:
            sourcetype = REQUEST.get('sourcetype', sourcetype)
        if sourcetype == 'sourceFile':
            if REQUEST:
                xml_files = [REQUEST.get("file")]
            else:
                xml_files = [xmlstring]
        elif sourcetype == 'folder':
            if REQUEST:
                from_folder=REQUEST.get("from_folder")
            xml_files = glob.glob(os.path.join(from_folder, '*.xml'))

        files_counter = 0
        errors = 0
        imports = 0

        for xml_file in xml_files:
            if isinstance(xml_file, basestring):
                xmlstring = xml_file
            else:
                if hasattr(xml_file, 'read'):
                    xmlstring = xml_file.read()
                else:
                    fileobj = codecs.open(xml_file, 'r', 'utf-8')
                    xmlstring = fileobj.read().encode('utf-8')

            xmldoc = parseString(xmlstring)
            documents = xmldoc.getElementsByTagName("document")
            total_docs = len(documents)
            if total_docs > 1:
                logger.info("Documents count: %d" % total_docs)
            docs_counter = 0
            files_counter = files_counter + 1

            for d in documents:
                docid = d.getAttribute('id')
                try:
                    if self.documents.has_key(docid):
                        self.documents._delOb(docid)
                    self.importDocumentFromXML(d)
                    imports = imports + 1
                except PlominoReplicationException, e:
                    logger.info('error while importing %s (%s)' % (docid, e))
                    errors = errors + 1
                docs_counter = docs_counter + 1
                if docs_counter == 100:
                    self.setStatus("Importing documents (%d%%)" % int(100*docs_counter/total_docs))
                    txn.savepoint(optimistic=True)
                    docs_counter = 0
                    logger.info("%d documents imported successfully, %d errors(s) ...(still running)" % (imports, errors))
            if files_counter == 100:
                self.setStatus("Importing documents (%s)" % files_counter)
                txn.savepoint(optimistic=True)
                files_counter = 0
                logger.info("%d documents imported successfully, %d errors(s) ...(still running)" % (imports, errors))

        self.setStatus("Ready")
        logger.info("Importation finished: %d documents imported successfully, %d document(s) not imported" % (imports, errors))
        txn.commit()

        return (imports, errors)

    security.declareProtected(CREATE_PERMISSION, 'importDocumentFromXML')
    def importDocumentFromXML(self, node):
        docid = node.getAttribute('id')
        lastmodified = DateTime(node.getAttribute('lastmodified'))
        pt = getToolByName(self, 'portal_types')
        pt.constructContent('PlominoDocument', self.documents, docid)
        #self.invokeFactory(type_name='PlominoDocument', id=docid)
        doc = self.documents.get(docid)

        # restore items
        itemnode = node.getElementsByTagName("params")[0]
        #result, method = xmlrpclib.loads(node.firstChild.toxml())
        result, method = xmlrpclib.loads(itemnode.toxml().encode('utf-8'))
        items = result[0]
        for k in items.keys():
            # convert xmlrpclib.DateTime into DateTime
            if items[k].__class__.__name__=='DateTime':
                items[k]=StringToDate(items[k].value[:19], format="%Y-%m-%dT%H:%M:%S")
        doc.items = items

        # restore files
        for fnode in node.getElementsByTagName("attachment"):
            filename = str(fnode.getAttribute('id'))
            contenttype = str(fnode.getAttribute('contenttype'))
            doc.setfile(fnode.firstChild.data.decode('base64'), filename=filename, overwrite=True, contenttype=contenttype)
        doc.save(onSaveEvent=False)
        doc.plomino_modification_time = lastmodified




# -*- coding: utf-8 -*-
#
# File: googlechart.py
#
# Copyright (c) 2009 by ['Eric BREHAULT']
#
# Zope Public License (ZPL)
#

__author__ = """Eric BREHAULT <eric.brehault@makina-corpus.com>"""
__docformat__ = 'plaintext'

from zope.formlib import form
from zope.interface import implements
from zope.schema import getFields
from zope.schema import TextLine

from dictionaryproperty import DictionaryProperty

from base import IBaseField, BaseField, BaseForm

class IGooglechartField(IBaseField):
    """
    Google chart field schema
    """
    editrows = TextLine(title=u'Rows',
                      description=u'Size of the editable text area',
                      default=u"6",
                      required=False)

class GooglechartField(BaseField):
    """
    """
    implements(IGooglechartField)

    def validate(self, submittedValue):
        """
        """
        errors=[]
        # no validation needed (we do not want to parse the GoogleChart param)
        return errors

    def processInput(self, submittedValue):
        """
        """
        lines = submittedValue.replace('\r', '').split('\n')
        params = {}
        for l in lines:
            if "=" in l:
                (key, value) = l.split('=')
            else:
                key = l
                value = ''
            params[key] = value
        return params


for f in getFields(IGooglechartField).values():
    setattr(GooglechartField, f.getName(), DictionaryProperty(f, 'parameters'))

class SettingForm(BaseForm):
    """
    """
    form_fields = form.Fields(IGooglechartField)




# -*- coding: utf-8 -*-
#
# File: PlominoAction.py
#
# Copyright (c) 2008 by ['Eric BREHAULT']
# Generator: ArchGenXML Version 2.0
#            http://plone.org/products/archgenxml
#
# Zope Public License (ZPL)
#

__author__ = """Eric BREHAULT <eric.brehault@makina-corpus.org>"""
__docformat__ = 'plaintext'

from AccessControl import ClassSecurityInfo
from Products.Archetypes.atapi import *
from zope.interface import implements
import interfaces

from Products.CMFDynamicViewFTI.browserdefault import BrowserDefaultMixin

from Products.CMFPlomino.config import *
from exceptions import PlominoScriptException

##code-section module-header #fill in your manual code here
from Products.Archetypes.public import *

ACTION_TYPES = [["OPENFORM", "Open form"], ["OPENVIEW", "Open view"], ["CLOSE", "Close"], ["SAVE", "Save"], ["PYTHON", "Python script"], ["REDIRECT", "Redirect formula"]]
ACTION_DISPLAY = [["LINK", "Link"], ["SUBMIT", "Submit button"], ["BUTTON", "Button"]]
##/code-section module-header

schema = Schema((

    StringField(
        name='id',
        widget=StringField._properties['widget'](
            label="Id",
            description="The action id",
            label_msgid='CMFPlomino_label_action_id',
            description_msgid='CMFPlomino_help_action_id',
            i18n_domain='CMFPlomino',
        ),
    ),
    StringField(
        name='ActionType',
        widget=SelectionWidget(
            label="Action type",
            description="Select the type for this action",
            label_msgid='CMFPlomino_label_ActionType',
            description_msgid='CMFPlomino_help_ActionType',
            i18n_domain='CMFPlomino',
        ),
        vocabulary=ACTION_TYPES,
    ),
    StringField(
        name='ActionDisplay',
        widget=SelectionWidget(
            label="Action display",
            description="How the action is shown",
            label_msgid='CMFPlomino_label_ActionDisplay',
            description_msgid='CMFPlomino_help_ActionDisplay',
            i18n_domain='CMFPlomino',
        ),
        vocabulary=ACTION_DISPLAY,
    ),
    TextField(
        name='Content',
        widget=TextAreaWidget(
            label="Parameter or code",
            description="Code or parameter depending on the action type",
            label_msgid='CMFPlomino_label_ActionContent',
            description_msgid='CMFPlomino_help_ActionContent',
            i18n_domain='CMFPlomino',
        ),
    ),
    TextField(
        name='Hidewhen',
        widget=TextAreaWidget(
            label="Hide-when formula",
            description="Action is hidden if formula return True",
            label_msgid='CMFPlomino_label_ActionHidewhen',
            description_msgid='CMFPlomino_help_ActionHidewhen',
            i18n_domain='CMFPlomino',
        ),
    ),
    BooleanField(
        name='InActionBar',
        default="1",
        widget=BooleanField._properties['widget'](
            label="Display action in action bar",
            description="Display action in action bar (yes/no)",
            label_msgid='CMFPlomino_label_ActionInActionBar',
            description_msgid='CMFPlomino_help_ActionInActionBar',
            i18n_domain='CMFPlomino',
        ),
    ),
),
)

##code-section after-local-schema #fill in your manual code here
##/code-section after-local-schema

PlominoAction_schema = BaseSchema.copy() + \
    schema.copy()

##code-section after-schema #fill in your manual code here
##/code-section after-schema

class PlominoAction(BaseContent, BrowserDefaultMixin):
    """
    """
    security = ClassSecurityInfo()
    implements(interfaces.IPlominoAction)

    meta_type = 'PlominoAction'
    _at_rename_after_creation = False

    schema = PlominoAction_schema

    ##code-section class-header #fill in your manual code here
    ##/code-section class-header

    # Methods

    security.declareProtected(READ_PERMISSION, 'executeAction')
    def executeAction(self, target, form_id):
        """return the action resulting url
        """
        db = self.getParentDatabase()
        if self.ActionType == "OPENFORM":
            formid = self.Content()
            return db.absolute_url() + '/' + formid + '/OpenForm'
        elif self.ActionType == "OPENVIEW":
            viewid = self.Content()
            return db.absolute_url() + '/' + viewid + '/OpenView'
        elif self.ActionType == "CLOSE":
            return db.absolute_url() + '/checkBeforeOpenDatabase'
        elif self.ActionType == "PYTHON":
            if target is None:
                targetid="None"
            else:
                targetid=target.id
            return self.absolute_url() + '/runScript?target=%s&form_id=%s'%(targetid, form_id)
        elif self.ActionType == "REDIRECT":
            try:
                redirecturl=self.runFormulaScript("action_"+self.getParentNode().id+"_"+self.id+"_script", target, self.Content, True, form_id)
                return str(redirecturl)
            except PlominoScriptException, e:
                return "javascript:alert(\"Error: %s\")" % ('formula error in redirect action ' + self.Title())
        else:
            return '.'

    security.declareProtected(READ_PERMISSION, 'runScript')
    def runScript(self, REQUEST):
        """execute the python code
        """
        db = self.getParentDatabase()
        target = REQUEST.get('target')
        form_id = REQUEST.get('form_id')
        if target == "None":
            plominoContext = db
        else:
            plominoContext = db.getDocument(target)
            if plominoContext is None:
                plominoContext = getattr(db, target, db)

        plominoReturnURL = plominoContext.absolute_url()
        try:
            #RunFormula(plominoContext, "action_"+self.getParentNode().id+"_"+self.id, self.Content())
            returnurl=self.runFormulaScript("action_"+self.getParentNode().id+"_"+self.id+"_script", plominoContext, self.Content, True, form_id)
            if returnurl is None or returnurl=='':
                returnurl=plominoReturnURL
            REQUEST.RESPONSE.redirect(returnurl)
        except PlominoScriptException, e:
            e.reportError('"%s" action failed' % self.Title())
            REQUEST.RESPONSE.redirect(plominoReturnURL)

    security.declarePublic('at_post_edit_script')
    def at_post_edit_script(self):
        """post edit
        """
        self.cleanFormulaScripts("action_"+self.getParentNode().id+"_"+self.id)


registerType(PlominoAction, PROJECTNAME)
# end of class PlominoAction

##code-section module-footer #fill in your manual code here
##/code-section module-footer






# -*- coding: utf-8 -*-
#
# File: attachment.py
#
# Copyright (c) 2009 by ['Eric BREHAULT']
#
# Zope Public License (ZPL)
#

__author__ = """Eric BREHAULT <eric.brehault@makina-corpus.com>"""
__docformat__ = 'plaintext'

from zope.formlib import form
from zope.interface import implements
from zope.schema import getFields

from dictionaryproperty import DictionaryProperty


from base import IBaseField, BaseField

class IAttachmentField(IBaseField):
    """
    Number field schema
    """

class AttachmentField(BaseField):
    """
    """
    implements(IAttachmentField)


for f in getFields(IAttachmentField).values():
    setattr(AttachmentField, f.getName(), DictionaryProperty(f, 'parameters'))



# -*- coding: utf-8 -*-
#
# File: PlominoAgent.py
#
# Copyright (c) 2008 by ['Eric BREHAULT']
# Generator: ArchGenXML Version 2.0
#            http://plone.org/products/archgenxml
#
# Zope Public License (ZPL)
#

__author__ = """Eric BREHAULT and Xavier PERROT <eric.brehault@makina-corpus.org>"""
__docformat__ = 'plaintext'

from AccessControl import ClassSecurityInfo
from AccessControl.SecurityManagement import newSecurityManager
from Products.Archetypes.atapi import *
from zope.interface import implements
import interfaces
from zope.component import getUtility
try:
    from plone.app.async.interfaces import IAsyncService
    ASYNC = True
except:
    ASYNC = False
    
from Products.CMFDynamicViewFTI.browserdefault import BrowserDefaultMixin
from exceptions import PlominoScriptException
from Products.CMFPlomino.config import *

schema = Schema((

    StringField(
        name='id',
        widget=StringField._properties['widget'](
            label="Id",
            description="The agent id",
            label_msgid='CMFPlomino_label_agent_id',
            description_msgid='CMFPlomino_help_agent_id',
            i18n_domain='CMFPlomino',
        ),
    ),
    TextField(
        name='Content',
        widget=TextAreaWidget(
            label="Code",
            description="Code to execute",
            label_msgid="CMFPlomino_label_AgentContent",
            description_msgid="CMFPlomino_help_AgentContent",
            i18n_domain='CMFPlomino',
            rows=25,
        ),
    ),
    StringField(
        name='RunAs',
        default="CURRENT",
        widget=SelectionWidget(
            label="Run as",
            description="Run the agent using current user access rights, or using the developer access rights.",
            label_msgid='CMFPlomino_label_AgentRunAs',
            description_msgid='CMFPlomino_help_AgentRunAs',
            i18n_domain='CMFPlomino',
        ),
        vocabulary= [["CURRENT", "Current user"], ["OWNER", "Agent owner"]],
    ),
),
)

PlominoAgent_schema = BaseSchema.copy() + \
    schema.copy()
    
def run_async(context):
    # for async call
    context.runAgent()

class PlominoAgent(BaseContent, BrowserDefaultMixin):
    """
    """
    security = ClassSecurityInfo()
    implements(interfaces.IPlominoAgent)

    meta_type = 'PlominoAgent'
    _at_rename_after_creation = False

    schema = PlominoAgent_schema

    security.declarePublic('at_post_edit_script')
    def at_post_edit_script(self):
        """
        """
        self.cleanFormulaScripts("agent_"+self.id)

    def __call__(self, *args):
        """
        """
        plominoContext = self
        try:
            if self.getRunAs() == "OWNER":
                owner = self.getOwner()
                # user = self.getCurrentUser()
                newSecurityManager(None, owner)

            result = self.runFormulaScript("agent_"+self.id, plominoContext, self.Content, True, *args)
            # if self.getRunAs() == "OWNER":
            #     newSecurityManager(None, user)
        except PlominoScriptException, e:
            e.reportError('Agent failed')
            result = None

        return result

    security.declarePublic('runAgent')
    def runAgent(self, REQUEST=None):
        """ Execute the agent formula.
        """
        plominoContext = self
        plominoReturnURL = self.getParentDatabase().absolute_url()
        if not REQUEST:
            REQUEST = getattr(self, 'REQUEST', None)
        try:
            if self.getRunAs() == "OWNER":
                user = self.getOwner()
                newSecurityManager(None, user)

            r=self.runFormulaScript("agent_"+self.id, plominoContext, self.Content)
            if (REQUEST != None) and (REQUEST.get('REDIRECT', None)== "True"):
                if r is not None:
                    plominoReturnURL=r
                REQUEST.RESPONSE.redirect(plominoReturnURL)
        except PlominoScriptException, e:
            if REQUEST:
                REQUEST.RESPONSE.setHeader('content-type', 'text/plain; charset=utf-8')
            return e.message

    security.declarePublic('runAgent_async')
    def runAgent_async(self):
        """run the agent in asynchronous mode
        """
        async = getUtility(IAsyncService)
        job = async.queueJob(run_async, self)
        
registerType(PlominoAgent, PROJECTNAME)
# end of class PlominoAgent

##code-section module-footer #fill in your manual code here
##/code-section module-footer






from zope.interface import directlyProvides
from Products.BTreeFolder2.BTreeFolder2 import manage_addBTreeFolder
from Products.CMFPlone.interfaces import IHideFromBreadcrumbs

from Products.CMFPlomino.fields.text import ITextField
from Products.CMFPlomino.fields.selection import ISelectionField
from Products.CMFPlomino.fields.number import INumberField
from Products.CMFPlomino.fields.datetime import IDatetimeField
from Products.CMFPlomino.fields.name import INameField
from Products.CMFPlomino.fields.doclink import IDoclinkField

from Products.CMFPlomino.PlominoUtils import asUnicode

from Products.PythonScripts.PythonScript import PythonScript
from Products.PythonScripts.PythonScript import manage_addPythonScript
from OFS.Image import manage_addImage
import parser
import re

import logging
logger = logging.getLogger('Plomino migration')

def migrate(db):
    """
    """
    messages = []
    if not(hasattr(db, "plomino_version")):
        msg = migrate_to_130(db)
        messages.append(msg)
    if db.plomino_version=="1.3.0":
        # no migration needed here
        db.plomino_version = "1.4.0"
    if db.plomino_version=="1.4.0":
        msg = migrate_to_15(db)
        messages.append(msg)
    if db.plomino_version=="1.5":
        msg = migrate_to_16(db)
        messages.append(msg)
    if db.plomino_version=="1.6":
        msg = migrate_to_161(db)
        messages.append(msg)
    if db.plomino_version=="1.6.1":
        # no migration needed here
        db.plomino_version = "1.6.2"
    if db.plomino_version=="1.6.2":
        # no migration needed here
        db.plomino_version = "1.6.3"
    if db.plomino_version=="1.6.3":
        msg = migrate_to_164(db)
        messages.append(msg)
    if db.plomino_version=="1.6.4":
        msg = migrate_to_17(db)
        messages.append(msg)
    if db.plomino_version=="1.7":
        msg = migrate_to_173(db)
        messages.append(msg)
    if db.plomino_version=="1.7.3":
        # no migration needed here
        db.plomino_version = "1.7.4"
    if db.plomino_version=="1.7.4":
        msg = migrate_to_175(db)
        messages.append(msg)
    if db.plomino_version=="1.7.5":
        # no migration needed here
        db.plomino_version = "1.8"
    if db.plomino_version=="1.8":
        # no migration needed here
        db.plomino_version = "1.9"
    if db.plomino_version=="1.9":
        msg = migrate_to_1_10(db)
        messages.append(msg)
    if db.plomino_version=="1.10":
        # no migration needed here
        db.plomino_version = "1.10.3"
    if db.plomino_version=="1.10.3":
        msg = migrate_to_1_11(db)
        messages.append(msg)
    if db.plomino_version=="1.11":
        msg = migrate_to_1_12(db)
        messages.append(msg)
    if db.plomino_version=="1.12":
        # no migration needed here
        db.plomino_version = "1.13"
    return messages

def migrate_to_130(db):
    """ PlominoField schema has been changed, all type-specific parameters
    have been removed and are now handled through specific adapters.
    """
    db.setDebugMode(False)
    msg = "Migration to 1.3: DebugMode attribute added"

    for form in db.getForms():
        for field in form.getFormFields():
            type = field.getFieldType()
            if type == "TEXT":
                adapt = ITextField(field)
                adapt.widget = "TEXT"
            elif type == "NUMBER":
                adapt = INumberField(field)
                adapt.type = "INTEGER"
            elif type == "FLOAT":
                field.setFieldType("NUMBER")
                adapt = INumberField(field)
                adapt.type = "FLOAT"
            elif type == "NAME":
                adapt = INameField(field)
                adapt.type = "SINGLE"
            elif type == "NAMES":
                field.setFieldType("NAME")
                adapt = INameField(field)
                adapt.type = "MULTI"
            elif type in ("SELECTION", "MULTISELECTION", "CHECKBOX", "RADIO"):
                field.setFieldType("SELECTION")
                adapt = ISelectionField(field)
                l = getattr(field, "SelectionList", None)
                if l is not None:
                    adapt.selectionlist = list(l)
                v = getattr(field, "SelectionListFormula", None)
                if v is not None:
                    adapt.selectionlistformula = v.raw
                adapt.separator = getattr(field, "DisplayModList", getattr(field, "OtherDisplayMod", None))
                if type == "SELECTION":
                    adapt.widget = "SELECT"
                elif type == "MULTISELECTION":
                    adapt.widget = "MULTISELECT"
                elif type == "CHECKBOX":
                    adapt.widget = "CHECKBOX"
                elif type == "RADIO":
                    adapt.widget = "RADIO"
            elif type == "DOCLINK":
                adapt = IDoclinkField(field)
                v = getattr(field, "SelectionListFormula", None)
                if v is not None:
                    adapt.documentslistformula = v.raw
    msg = msg + ", FieldType remapped" 
    db.plomino_version = "1.3.0"
    return msg

def migrate_to_15(db):
    """ new attribute in Column: DisplaySum
    """
    for v_obj in db.getViews():
        for c in v_obj.getColumns():
            c.setDisplaySum(False)
    msg = "Migration to 1.5: DisplaySum attribute added"
    db.plomino_version = "1.5"
    return msg

def migrate_to_16(db):
    """ attribute Position in column is replaced by AtFolder sorting
    """
    for v_obj in db.getViews():
        # sort columns by their Position
        orderedcolumns = []
        for c in v_obj.getColumns():
            if not(c is None):
                orderedcolumns.append([c.Position, c])
        orderedcolumns.sort()

        # set the position using the previous sorting
        for i, c in enumerate(orderedcolumns):
            v_obj.moveObject(c[1].id, i)
            v_obj.plone_utils.reindexOnReorder(v_obj)

    msg = "Migration to 1.6: Position column attribute deleted"
    db.plomino_version = "1.6"
    return msg

def migrate_to_161(db):
    """ Fulltext index db attribute.
    """
    db.setFulltextIndex(False)
    msg = "Migration to 1.6.1: Fulltext index db attribute set to False"
    db.plomino_version = "1.6.1"
    return msg

def migrate_to_164(db):
    """ Fulltext index db attribute.
    """
    db.setDoNotReindex(False)
    msg = "Migration to 1.6.4: DoNotReindex db attribute set to False"
    db.plomino_version = "1.6.4"
    return msg

def migrate_to_17(db):
    """ dynamic hidewhen attribute
    """
    #for form in db.getForms():
    #    form.setIsDynamicHidewhen(False)
    msg = "Migration to 1.7: Dynamic hide-when initialized"
    db.plomino_version = "1.7"
    return msg

def migrate_to_173(db):
    """ dynamic hidewhen attribute moved
    """
    for form in db.getForms():
        for hidewhen in form.getHidewhenFormulas():
            hidewhen.setIsDynamicHidewhen(False)
    msg = "Migration to 1.7.3: Dynamic hide-when initialized"
    db.plomino_version = "1.7.3"
    return msg

def migrate_to_175(db):
    """ documents stores in BTreeFolder
    """
    manage_addBTreeFolder(db, id='plomino_documents')
    directlyProvides(db.plomino_documents, IHideFromBreadcrumbs)
    docids = [id for id in db.objectIds() if getattr(db, id).portal_type == "PlominoDocument"]
    cookie = db.manage_cutObjects(ids=docids)
    db.plomino_documents.manage_pasteObjects(cookie)
    msg = "Migration to 1.7.5: Documents moved in BTreeFolder"
    db.plomino_version = "1.7.5"
    return msg

def migrate_to_1_10(db):
    """ rename Plomino_Portlet_Availability fields
    """
    for form in db.getForms():
        for field in form.getFormFields():
            if field.id == "Plomino_Portlet_Availabilty":
                form.manage_renameObject("Plomino_Portlet_Availabilty", "Plomino_Portlet_Availability")
                field.setTitle("Plomino_Portlet_Availability")
    msg = "Migration to 1.10: Rename Plomino_Portlet_Availability fields"
    db.plomino_version = "1.10"
    return msg

def migrate_to_1_11(db):
    """ getAllDocuments now returns documents, unless requested.
    """
    # # Field formulas
    # # Selection formulas
    # # Hidewhen formulas
    # getHidewhenFormulas
    # # Column formulas
    # # Agents
    # agent.Content
    # # Script libraries
    # File: use str() to get the content and manage_edit to set it.
    # Page Template: read() and write()
    from zope.interface import providedBy
    from Products.CMFPlomino.fields.selection import ISelectionField
    
    forms = db.getForms()
    for form in forms:
        fields = form.getFormFields()
        for field in fields:
            # NB: Don't use getFormula, this strips markup from strings.
            f = field.Formula()
            if f:
                logger.info("Migrated formula: %s\n"
                            "Old version: %s"%(field, f))
                field.setFormula(
                    f.replace('getAllDocuments()', 'getAllDocuments(getObject=False)'))
            f = field.ValidationFormula()
            if f:
                logger.info("Migrated validation formula: %s\n"
                            "Old version: %s"%(field, f))
                field.setValidationFormula(
                    f.replace('getAllDocuments()', 'getAllDocuments(getObject=False)'))
            settings = field.getSettings()
            if ISelectionField in providedBy(settings).interfaces():
                selectionlistformula = settings.selectionlistformula
                if selectionlistformula:
                    settings.selectionlistformula = selectionlistformula.replace(
                        'getAllDocuments()', 'getAllDocuments(getObject=False)')
        hidewhens = form.getHidewhenFormulas()
        for hidewhen in hidewhens:
            f = hidewhen.Formula()
            if f:
                logger.info("Migrated hidewhen formula: %s\n"
                            "Old version: %s"%(hidewhen, f))
                hidewhen.setFormula(
                    f.replace('getAllDocuments()', 'getAllDocuments(getObject=False)'))
        actions = form.objectValues(spec='PlominoAction')
        for action in actions:
            f = action.Content()
            if f:
                logger.info("Migrated action formula: %s\n"
                            "Old version: %s"%(action, f))
                action.setContent(
                    f.replace('getAllDocuments()', 'getAllDocuments(getObject=False)'))
            f = action.Hidewhen()
            if f:
                logger.info("Migrated action hidewhen: %s\n"
                            "Old version: %s"%(action, f))
                action.setHidewhen(
                    f.replace('getAllDocuments()', 'getAllDocuments(getObject=False)'))
    views = db.getViews()
    for view in views:
        columns = view.getColumns()
        for column in columns:
            f = column.Formula()
            if f:
                logger.info("Migrated column formula: %s\n"
                            "Old version: %s"%(column, f))
                column.setFormula(
                    f.replace('getAllDocuments()', 'getAllDocuments(getObject=False)'))
        actions = view.objectValues(spec='PlominoAction')
        for action in actions:
            f = action.Content()
            if f:
                logger.info("Migrated action formula: %s\n"
                            "Old version: %s"%(action, f))
                action.setContent(
                    f.replace('getAllDocuments()', 'getAllDocuments(getObject=False)'))
            f = action.Hidewhen()
            if f:
                logger.info("Migrated action hidewhen: %s\n"
                            "Old version: %s"%(action, f))
                action.setHidewhen(
                    f.replace('getAllDocuments()', 'getAllDocuments(getObject=False)'))
    agents = db.getAgents()
    for agent in agents:
        f = agent.Content()
        if f:
            logger.info("Migrated agent formula: %s\n"
                        "Old version: %s"%(agent, f))
            agent.setContent(
                f.replace('getAllDocuments()', 'getAllDocuments(getObject=False)'))
    files = db.resources.objectValues('File')
    for f in files:
        if f.content_type.startswith('text'):
            formula = asUnicode(f).encode('utf-8')
            logger.info("Migrated script library formula: %s"%f.id())
            f.manage_edit(f.title, f.content_type, filedata=formula.replace(
                'getAllDocuments()', 'getAllDocuments(getObject=False)'))
    templates = db.resources.objectValues('Page Template')
    for template in templates:
        f = template.read()
        logger.info("Migrated template formula: %s"%template.id)
        template.write(
          f.replace('getAllDocuments()', 'getAllDocuments(getObject=False)'))

    msg = "Migration to 1.11: getAllDocuments API change."
    db.plomino_version = "1.11"
    return msg

def migrate_to_1_12(db):
    """ Convert resources script lib File into PythonScript and Image
    """
    libs = db.resources.objectValues('File')
    for lib in libs:
        lib_id = lib.id()
        lib_data = lib.data
        content_type = lib.getContentType() 
        if 'image' in content_type:
            db.resources.manage_delObjects(lib_id)
            lib_id = manage_addImage(db.resources, lib_id, lib_data)
        else:
            error_re = re.compile('^## Errors:$', re.MULTILINE)
            ps = PythonScript('testing')
            try:
                lib_data = asUnicode(lib_data)
            except UnicodeDecodeError, e:
                logger.info("Unknown encoding, skipping: %s" % lib_id)
                continue
            
            ps.write(lib_data)
            if not error_re.search(ps.read()):
                db.resources.manage_delObjects(lib_id)
                blank = manage_addPythonScript(db.resources, lib_id)
                sc = db.resources._getOb(lib_id)
                sc.write(lib_data)
                logger.info("Converted to Script: %s" % lib_id)
                continue
    
    msg = "Migration to 1.12: Convert resources script lib File into PythonScripts."
    db.plomino_version = "1.12"
    return msg



# -*- coding: utf-8 -*-

from zope.interface import Interface

class IPlominoDatabase(Interface):
    """Marker interface for .PlominoDatabase.PlominoDatabase
    """

class IPlominoAction(Interface):
    """Marker interface for .PlominoAction.PlominoAction
    """

class IPlominoForm(Interface):
    """Marker interface for .PlominoForm.PlominoForm
    """

class IPlominoField(Interface):
    """Marker interface for .PlominoField.PlominoField
    """

class IPlominoView(Interface):
    """Marker interface for .PlominoView.PlominoView
    """

class IPlominoColumn(Interface):
    """Marker interface for .PlominoColumn.PlominoColumn
    """

class IPlominoDocument(Interface):
    """Marker interface for .PlominoDocument.PlominoDocument
    """

class IPlominoHidewhen(Interface):
    """Marker interface for .PlominoHidewhen.PlominoHidewhen
    """

class IPlominoAgent(Interface):
    """Marker interface for .PlominoAgent.PlominoAgent
    """

class IPlominoUtils(Interface):
    """Marker interface for PlominoUtils
    """

class IPlominoCache(Interface):
    """Marker interface for PlominoCache
    """



# -*- coding: utf-8 -*-
#
# File: PlominoCache.py
#
# Copyright (c) 2011 by ['Eric BREHAULT']
# Zope Public License (ZPL)
#

__author__ = """Eric BREHAULT <eric.brehault@makina-corpus.org>"""
__docformat__ = 'plaintext'

from AccessControl import ClassSecurityInfo
from Products.Archetypes.atapi import *
from zope.interface import implements
import interfaces

from Products.CMFDynamicViewFTI.browserdefault import BrowserDefaultMixin

from Products.CMFPlomino.config import *

schema = Schema((

    StringField(
        name='id',
        widget=StringField._properties['widget'](
            label="Id",
            description="The cache fragment id",
            label_msgid='CMFPlomino_label_cache_id',
            description_msgid='CMFPlomino_help_cache_id',
            i18n_domain='CMFPlomino',
        ),
    ),
    TextField(
        name='Formula',
        widget=TextAreaWidget(
            label="Formula",
            description="Returns the cache key (if None, no cache)",
            label_msgid='CMFPlomino_label_CacheFormula',
            description_msgid='CMFPlomino_help_CacheFormula',
            i18n_domain='CMFPlomino',
        ),
    ),
),
)

PlominoCache_schema = BaseSchema.copy() + \
    schema.copy()

class PlominoCache(BaseContent, BrowserDefaultMixin):
    """
    """
    security = ClassSecurityInfo()
    implements(interfaces.IPlominoCache)

    meta_type = 'PlominoCache'
    _at_rename_after_creation = False

    schema = PlominoCache_schema

    security.declarePublic('at_post_edit_script')
    def at_post_edit_script(self):
        """
        """
        self.cleanFormulaScripts("cache_"+self.getParentNode().id+"_"+self.id)


registerType(PlominoCache, PROJECTNAME)


import traceback

from PlominoUtils import asUnicode
from Products.CMFCore.utils import getToolByName

import logging
logger = logging.getLogger('Plomino')

class PlominoScriptException(Exception):

    def __init__(self, context, exception_obj, formula, script_id):
        """
        """
        self.context = context
        self.context_url = context.absolute_url_path()
        self.formula = formula
        self.script_id = script_id
        self.message = asUnicode(self.traceErr())

    def __str__(self):
        """
        """
        msg = self.message.encode('utf-8')
        return msg

    def traceErr(self):
        """trace errors from Scripts
        """
        msg = []

        formatted_lines = traceback.format_exc().splitlines()
        index = 0
        for (i, line) in enumerate(formatted_lines):
            if self.script_id in line:
                index = i
        if index == len(formatted_lines)-1:
            msg = [formatted_lines[index]]
        else:
            msg = [formatted_lines[index], formatted_lines[-1]]
        msg.append("Context is <%s> %s" % (self.context.__class__.__name__, self.context_url))
        if self.context.getParentDatabase().debugMode:
            code = ["Code : "]
            line_number = 4
            for l in self.formula().replace('\r', '').split('\n'):
                code.append("%d: %s" % (line_number, l))
                line_number += 1
            logger.error('\n'.join(msg+code))
            
        return "\r\n".join(msg)

    def reportError(self, label, path=None, request=None):
        """
        """
        report = asUnicode(label)
        if not request:
            request = getattr(self.context, 'REQUEST', None)
        if request:
            if not path:
                if self.formula and hasattr(self.formula, 'absolute_url_path'):
                    path = self.formula.absolute_url_path()
            if path:
                report += " - Plomino formula %s" % path

            traceback = self.message.replace("<", "&lt;").replace(">", "&gt;")
            report += " - Plomino traceback " + traceback.replace('\n', '\n<br/>')
            plone_tools = getToolByName(self.context.getParentDatabase().aq_inner, 'plone_utils')
            plone_tools.addPortalMessage(report, 'error', request)
            
class PlominoRenderingException(Exception):
    pass

class PlominoReplicationException(Exception):
    pass

class PlominoDesignException(Exception):
    pass

class PlominoCacheException(Exception):
    pass


# -*- coding: utf-8 -*-
#
# File: dictionaryproperty.py
#
# Copyright (c) 2008 by ['Eric BREHAULT']
#
# Zope Public License (ZPL)
#
_marker = object()

class DictionaryProperty(object):
    """Computed attributes based on schema fields stored in a dictionary
    (based on zope/schema/fieldproperty.py) 
    """

    def __init__(self, field, dictionary, name=None):
        if name is None:
            name = field.__name__

        self.__field = field
        self.__name = name
        self.__dictionary = dictionary

    def __get__(self, inst, klass):
        if inst is None:
            return self

        value = getattr(inst, self.__dictionary).get(self.__name, _marker)
        if value is _marker:
            field = self.__field.bind(inst)
            value = getattr(field, 'default', _marker)
            if value is _marker:
                raise AttributeError(self.__name)

        return value

    def __set__(self, inst, value):
        field = self.__field.bind(inst)
        field.validate(value)
        if field.readonly and inst.__dict__.has_key(self.__name):
            raise ValueError(self.__name, 'field is readonly')
        getattr(inst, self.__dictionary)[self.__name] = value

    def __getattr__(self, name):
        return getattr(self.__field, name)



# -*- coding: utf-8 -*-
#
# File: PlominoDocument.py
#
# Copyright (c) 2008 by ['Eric BREHAULT']
# Generator: ArchGenXML Version 2.0
#            http://plone.org/products/archgenxml
#
# Zope Public License (ZPL)
#

__author__ = """Eric BREHAULT <eric.brehault@makina-corpus.org>"""
__docformat__ = 'plaintext'

from AccessControl import ClassSecurityInfo
from Products.Archetypes.atapi import *
from zope.interface import implements
import interfaces
from Products.ATContentTypes.content.folder import ATFolder
from Products.CMFDynamicViewFTI.browserdefault import BrowserDefaultMixin
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import normalizeString

from exceptions import PlominoScriptException
from Products.CMFPlomino.config import *

from interfaces import *
from Products.Archetypes.public import *
from AccessControl import Unauthorized
from time import strptime
from DateTime import DateTime
from zope import event
from Products.Archetypes.event import ObjectEditedEvent
from zope.component import queryUtility

import simplejson as json

import logging
logger = logging.getLogger('Plomino')

# Import conditionally, so we don't introduce a hard dependency
try:
    from plone.i18n.normalizer.interfaces import IUserPreferredURLNormalizer
    from plone.i18n.normalizer.interfaces import IURLNormalizer
    URL_NORMALIZER = True
except ImportError:
    URL_NORMALIZER = False
    
from PlominoUtils import DateToString, StringToDate, sendMail, asUnicode, asList, PlominoTranslate
from OFS.Image import File
from ZPublisher.HTTPRequest import FileUpload
try:
    from iw.fss.FileSystemStorage import FileSystemStorage, FSSFileInfo
except Exception, e:
    pass
try:
    from plone.app.blob.field import BlobWrapper
    from plone.app.blob.utils import guessMimetype
    HAS_BLOB = True
except Exception, e:
    HAS_BLOB = False

schema = Schema((

),
)

##code-section after-local-schema #fill in your manual code here
##/code-section after-local-schema

PlominoDocument_schema = getattr(ATFolder, 'schema', Schema(())).copy() + \
    schema.copy()

##code-section after-schema #fill in your manual code here
##/code-section after-schema

class PlominoDocument(ATFolder):
    """
    """
    security = ClassSecurityInfo()
    implements(interfaces.IPlominoDocument)

    meta_type = 'PlominoDocument'
    _at_rename_after_creation = False

    schema = PlominoDocument_schema

    ##code-section class-header #fill in your manual code here
    ##/code-section class-header

    # Methods

    security.declarePublic('__init__')
    def __init__(self,oid,**kw):
        """initialization
        """
        ATFolder.__init__(self, oid, **kw)
        self.items={}
        self.plomino_modification_time = DateTime().toZone('UTC')

    security.declarePublic('checkBeforeOpenDocument')
    def checkBeforeOpenDocument(self):
        """check read permission and open view  NOTE: if READ_PERMISSION
        set on the 'view' action itself, it causes error 'maximum
        recursion depth exceeded' if user hasn't permission
        """
        if self.isReader():
            return self.OpenDocument()
        else:
            raise Unauthorized, "You cannot read this content"

    def doc_path(self):
        #db = self.getParentDatabase()
        #return db.getPhysicalPath() + (self.id,)
        # here we return actual path, we do not hide plomino_documents anymore
        return self.getPhysicalPath()

    def doc_url(self):
        """ return valid and nice url:
        - hide plomino_documents
        - use physicalPathToURL if REQUEST available
        """
        path = self.doc_path()
        short_path = [p for p in path if p!="plomino_documents"]
        if hasattr(self, "REQUEST"):
            return self.REQUEST.physicalPathToURL(short_path)
        else:
            return "/".join(short_path)
    
    security.declarePublic('setItem')
    def setItem(self,name,value):
        """
        """
        items = self.items
        if type(value) == type(''):
            db = self.getParentDatabase()
            translation_service = getToolByName(db, 'translation_service')
            value = translation_service.asunicodetype(value)
        items[name] = value
        self.items = items
        self.plomino_modification_time = DateTime().toZone('UTC')

    security.declarePublic('getItem')
    def getItem(self,name, default=''):
        """
        """
        if(self.items.has_key(name)):
            return self.items[name]
        else:
            return default

    security.declarePublic('hasItem')
    def hasItem(self,name):
        """
        """
        return self.items.has_key(name)

    security.declarePublic('removeItem')
    def removeItem(self,name):
        """
        """
        if(self.items.has_key(name)):
            items = self.items
            del items[name]
            self.items = items

    security.declarePublic('getItems')
    def getItems(self):
        """
        """
        return self.items.keys()

    security.declarePublic('getItemClassname')
    def getItemClassname(self,name):
        """
        """
        return self.getItem(name).__class__.__name__

    security.declarePublic('getLastModified')
    def getLastModified(self,asString=False):
        """
        """
        if not hasattr(self, 'plomino_modification_time'):
            self.plomino_modification_time = self.bobobase_modification_time().toZone('UTC')
        if asString:
            return str(self.plomino_modification_time)
        else:
            return self.plomino_modification_time

    security.declarePublic('getRenderedItem')
    def getRenderedItem(self, itemname, form=None, formid=None, convertattachments=False):
        """ return the item value rendered according the field defined in the given form
        (use default doc form if None)
        """
        result = ''
        db = self.getParentDatabase()
        if not form:
            if not formid:
                form = self.getForm()
            else:
                form = db.getForm(formid)
        if form:
            field = form.getFormField(itemname)
            if field:
                result = field.getFieldRender(form, self, False)
                if field.getFieldType()=='ATTACHMENT' and convertattachments:
                    result = result + ' ' + db.getIndex().convertFileToText(self,itemname)
                return result

        return result

    security.declarePublic('tojson')
    def tojson(self, REQUEST=None, item=None, formid=None):
        """return item value as JSON
        (return all items if item=None)
        """
        if not self.isReader():
            raise Unauthorized, "You cannot read this content"
        
        if REQUEST:
            REQUEST.RESPONSE.setHeader('content-type', 'application/json; charset=utf-8')
            item = REQUEST.get('item', item)
            formid = REQUEST.get('formid', formid)
        if not item:
            return json.dumps(self.items)
        
        if not formid:
            form = self.getForm()
        else:
            form = self.getParentDatabase().getForm(formid)
        if form:
            field = form.getFormField(item)
            if field:
                adapt = field.getSettings()
                fieldvalue = adapt.getFieldValue(form, self, False, False, REQUEST)
            else:
                fieldvalue = self.getItem(item)
        else:
            fieldvalue = self.getItem(item)
        return json.dumps(fieldvalue) 

    security.declarePublic('computeItem')
    def computeItem(self, itemname, form=None, formid=None, store=True, report=True):
        """ return the item value according the formula of the field defined in
        the given form (use default doc form if None)
        and store the value in the doc (if store=True)
        """
        result = None
        db = self.getParentDatabase()
        if not form:
            if not formid:
                form = self.getForm()
            else:
                form = db.getForm(formid)
        if form:
            result = form.computeFieldValue(itemname, self, report=report)
            if store:
                self.setItem(itemname, result)
        return result

    security.declarePublic('getPlominoReaders')
    def getPlominoReaders(self):
        """
        """
        if self.hasItem('Plomino_Readers'):
            return asList(self.Plomino_Readers)
        else:
            return ['*']

    security.declarePublic('isReader')
    def isReader(self):
        """
        """
        return self.getParentDatabase().isCurrentUserReader(self)
        
    security.declarePublic('isAuthor')
    def isAuthor(self):
        """
        """
        return self.getParentDatabase().isCurrentUserAuthor(self)

    security.declareProtected(REMOVE_PERMISSION, 'delete')
    def delete(self, REQUEST=None):
        """delete the current doc
        """
        db = self.getParentDatabase()
        db.deleteDocument(self)
        if not REQUEST is None:
            return_url = REQUEST.get('returnurl')
            REQUEST.RESPONSE.redirect(return_url)

    security.declareProtected(EDIT_PERMISSION, 'saveDocument')
    def saveDocument(self, REQUEST, creation=False):
        """save a document using the form submitted content
        """
        db = self.getParentDatabase()
        form = db.getForm(REQUEST.get('Form'))

        errors=form.validateInputs(REQUEST, doc=self)
        if len(errors)>0:
            return form.notifyErrors(errors)

        self.setItem('Form', form.getFormName())

        # process editable fields (we read the submitted value in the request)
        form.readInputs(self, REQUEST, process_attachments=True)

        # refresh computed values, run onSave, reindex
        self.save(form, creation)

        redirect = REQUEST.get('plominoredirecturl')
        if not redirect:
            redirect = self.getItem("plominoredirecturl")
        if not redirect:
            redirect = self.absolute_url()
        REQUEST.RESPONSE.redirect(redirect)


    security.declareProtected(EDIT_PERMISSION, 'refresh')
    def refresh(self, form=None):
        """ re-compute fields and re-index document
        (onSave event is not called, and authors are not updated 
        """
        self.save(form, creation=False, refresh_index=True, asAuthor=False, onSaveEvent=False)

    security.declareProtected(EDIT_PERMISSION, 'save')
    def save(self, form=None, creation=False, refresh_index=True, asAuthor=True, onSaveEvent=True):
        """refresh values according form, and reindex the document
        """
        # we process computed fields (refresh the value)
        # TODO: manage computed fields dependencies
        if form is None:
            form = self.getForm()
        else:
            self.setItem('Form', form.getFormName())

        db=self.getParentDatabase()
        if form:
            for f in form.getFormFields(includesubforms=True, doc=self, applyhidewhen=False):
                mode = f.getFieldMode()
                fieldname = f.id
                if mode in ["COMPUTED", "COMPUTEDONSAVE"] or (mode=="CREATION" and creation):
                    result = form.computeFieldValue(fieldname, self)
                    self.setItem(fieldname, result)
                else:
                    # computed for display field are not stored
                    pass

            # compute the document title
            try:
                result = self.runFormulaScript("form_"+form.id+"_title", self, form.DocumentTitle)
            except PlominoScriptException, e:
                e.reportError('Title formula failed')
                result = ""
            if not result:
                result = form.Title()
            self.setTitle(result)

            # update the document id
            if creation and form.getDocumentId():
                self._renameAfterCreation()
                # _renameAfterCreation index doc in portal_catalog
                # 1: we do not necessarily want it (depending on IndexInPortal value)
                # 2: we will index it with the correct path anyway
                # so let's remove it for now
                db.portal_catalog.uncatalog_object("/".join(self.getPhysicalPath()))
            
        # update the Plomino_Authors field with the current user name
        if asAuthor:
            authors = self.getItem('Plomino_Authors')
            name = db.getCurrentUser().getUserName()
            if authors == '':
                authors = []
                authors.append(name)
            elif name in authors:
                pass
            else:
                authors.append(name)
            self.setItem('Plomino_Authors', authors)

        # execute the onSaveDocument code of the form
        if form and onSaveEvent:
            try:
                result = self.runFormulaScript("form_"+form.id+"_onsave", self, form.onSaveDocument)
                if result and hasattr(self, 'REQUEST'):
                    self.REQUEST.set('plominoredirecturl', result)
            except PlominoScriptException, e:
                if hasattr(self, 'REQUEST'):
                    e.reportError('Document has been saved but onSave event failed.')
                    doc_path = self.REQUEST.physicalPathToURL(self.doc_path())
                    self.REQUEST.RESPONSE.redirect(doc_path)

        if refresh_index:
            # update index
            db.getIndex().indexDocument(self)
            # update portal_catalog
            if db.getIndexInPortal():
                db.portal_catalog.catalog_object(self, "/".join(db.getPhysicalPath() + (self.id,)))
            event.notify(ObjectEditedEvent(self))

    security.declareProtected(READ_PERMISSION, 'openWithForm')
    def openWithForm(self, form, editmode=False):
        """ Display the document using the given form's layouts.
        First, check if the user has proper access rights.
        """

        db = self.getParentDatabase()
        if editmode:
            if not db.isCurrentUserAuthor(self):
                raise Unauthorized, "You cannot edit this document."
        else:
            if not self.isReader():
                raise Unauthorized, "You cannot read this content"

        # execute the onOpenDocument code of the form
        valid = ''
        try:
            if form.getOnOpenDocument():
                #RunFormula(self, form.getOnOpenDocument())
                valid = self.runFormulaScript("form_"+form.id+"_onopen", self, form.onOpenDocument)
        except PlominoScriptException, e:
            e.reportError('onOpen event failed')

        if not valid:
            # we use the specified form's layout
            request = getattr(self, 'REQUEST', None)
            if not request:
                import sys
                from ZPublisher.HTTPResponse import HTTPResponse
                from ZPublisher.HTTPRequest import HTTPRequest
                response = HTTPResponse(stdout=sys.stdout)
                env = {'SERVER_NAME':'fake_server',
                       'SERVER_PORT':'80',
                       'REQUEST_METHOD':'GET'}
                request = HTTPRequest(sys.stdin, env, response)
            html_content = form.displayDocument(self,
                                editmode=editmode,
                                request=request)
        else:
            html_content = valid

        plone_tools = getToolByName(db, 'plone_utils')
        encoding = plone_tools.getSiteEncoding()
        html_content = html_content.encode(encoding) 

        return html_content


    security.declareProtected(EDIT_PERMISSION, 'editWithForm')
    def editWithForm(self,form):
        """
        """
        return self.openWithForm(form, True)

    security.declarePublic('send')
    def send(self,recipients,title,form=None):
        """Send current doc by mail
        """
        db = self.getParentDatabase()
        if form is None:
            form = self.getForm()
        message = self.openWithForm(form)
        sendMail(db, recipients, title, message)

    security.declarePublic('getForm')
    def getForm(self):
        """ Return form: look in REQUEST, then try to acquire from view, and
        finally fall back to document Form item.
        """
        formname = None
        if hasattr(self, 'REQUEST'):
            formname = self.REQUEST.get("openwithform", None)
        if not formname:
            if hasattr(self, 'evaluateViewForm'):
                formname = self.evaluateViewForm(self)
        if not formname:
            formname = self.getItem('Form')
        form = self.getParentDatabase().getForm(formname)
        if not form:
            if hasattr(self, "REQUEST") and formname:
                self.writeMessageOnPage("Form %s does not exist." % formname, self.REQUEST, True)
        return form

    security.declarePrivate('manage_afterClone')
    def manage_afterClone(self,item):
        """after cloning
        """
        # changed from BaseFolder to ATFolder because now inherits fron ATFolder
        ATFolder.manage_afterClone(self, item)

    security.declarePublic('__getattr__')
    def __getattr__(self, name):
        """Overloads getattr to return item values as attibutes
        """
        if(self.items.has_key(name)):
            return self.items[name]
        else:
            if name not in ['__parent__', '__conform__', '__annotations__',
                           '_v_at_subobjects', '__getnewargs__', 'aq_inner', 'im_self']:
                try:
                    return ATFolder.__getattr__(self, name)
                except Exception, e:
                    raise AttributeError, name
            else:   
                raise AttributeError, name

    security.declareProtected(READ_PERMISSION, 'isSelectedInView')
    def isSelectedInView(self,viewname):
        """
        """
        db = self.getParentDatabase()
        v = db.getView(viewname)
        result = False
        if v:
            try:
                #result = RunFormula(self, v.SelectionFormula())
                result = self.runFormulaScript("view_"+v.id+"_selection", self, v.SelectionFormula)
            except PlominoScriptException, e:
                e.reportError('%s view selection formula failed' % viewname)
        return result

    security.declareProtected(READ_PERMISSION, 'computeColumnValue')
    def computeColumnValue(self,viewname,columnname):
        """
        """
        db = self.getParentDatabase()
        v = db.getView(viewname)
        try:
            c = v.getColumn(columnname)
            #result = RunFormula(self, c.Formula())
            result = self.runFormulaScript("column_"+v.id+"_"+c.id+"_formula", self, c.Formula)
        except PlominoScriptException, e:
            e.reportError('"%s" column formula failed in %s view' % (c.Title(), viewname))
            result = None
        return result

    security.declareProtected(EDIT_PERMISSION, 'deleteAttachment')
    def deleteAttachment(self, REQUEST=None, fieldname=None, filename=None):
        """remove file object and update corresponding item value
        """
        if REQUEST is not None:
            fieldname=REQUEST.get('field')
            filename=REQUEST.get('filename')
        if fieldname is not None and filename is not None:
            current_files=self.getItem(fieldname)
            if current_files.has_key(filename):
                if len(current_files.keys()) == 1:
                    # if it is the only file attached, we need to make sure the field
                    # is not mandatory to allow deletion
                    form = self.getForm()
                    if form:
                        field = form.getFormField(fieldname)
                        if field and field.getMandatory():
                            error = fieldname + " " + PlominoTranslate("is mandatory", self)
                            return form.notifyErrors([error])
                del current_files[filename]
                self.setItem(fieldname, current_files)
                self.deletefile(filename)
        if REQUEST is not None:
            REQUEST.RESPONSE.redirect(self.absolute_url()+"/EditDocument")

    security.declarePublic('SearchableText')
    def SearchableText(self):
        values = []
        index_attachments=self.getParentDatabase().getIndexAttachments()
        form = self.getForm()

        for itemname in self.items.keys():
            item_value = self.getItem(itemname)
            if type(item_value) is list: 
                for v in item_value:
                    if type(v) is list:
                        values = values + [asUnicode(k) for k in v]
                    else:
                        values.append(asUnicode(v))
            else:
                values.append(asUnicode(item_value))
            # if selection or attachment field, we try to index rendered values too
            try:
                if form:
                    field = form.getFormField(itemname)
                    if field and field.getFieldType() in ["SELECTION", "ATTACHMENT"]:
                        v = asUnicode(self.getRenderedItem(itemname,form=form, convertattachments=index_attachments))
                        if v:
                            values.append(v)
            except:
                pass
        return ' '.join(values)

    security.declareProtected(READ_PERMISSION, 'getfile')
    def getfile(self, filename=None, REQUEST=None):
        """
        """
        if not self.isReader():
            raise Unauthorized, "You cannot read this content"

        fss = self.getParentDatabase().getStorageAttachments()
        if REQUEST is not None:
            filename = REQUEST.get('filename')
        if filename is not None:
            if fss:
                storage = FileSystemStorage()
                file_obj = storage.get(filename, self)
            else:
                #file_obj = getattr(self, filename)
                file_obj = self.get(filename, None)
            if not file_obj:
                return None
            if REQUEST is None:
                return file_obj
            else:
                REQUEST.RESPONSE.setHeader('content-type', file_obj.getContentType())
                REQUEST.RESPONSE.setHeader("Content-Disposition", "inline; filename="+filename)
                if fss:
                    return file_obj.getData()
                else:
                    return file_obj.data
        else:
            return None

    security.declareProtected(READ_PERMISSION, 'getFilenames')
    def getFilenames(self):
        """
        """
        fss = self.getParentDatabase().getStorageAttachments()
        if fss:
            return [o.getTitle() for o in self.__dict__.values() if isinstance(o, FSSFileInfo)]
        else:
            #return [o.getId() for o in self.getChildNodes() if isinstance(o, File)]
            if HAS_BLOB:
                return [id for id in self.objectIds() if isinstance(self[id], BlobWrapper)]
            else:
                return [id for id in self.objectIds() if isinstance(self[id], File)]

    security.declareProtected(EDIT_PERMISSION, 'setfile')
    def setfile(self, submittedValue, filename='', overwrite=False, contenttype=''):
        """
        """
        if filename=='':
            filename=submittedValue.filename
        if filename!='':
            if """\\""" in filename:
                filename=filename.split("\\")[-1]
            filename = ".".join([normalizeString(s, encoding='utf-8') for s in filename.split('.')])
            if filename in self.objectIds():
                if overwrite:
                    self.deletefile(filename)
                else:
                    return ("ERROR: "+filename+" already exists", "")
            if(self.getParentDatabase().getStorageAttachments()==True):
                tmpfile=File(filename, filename, submittedValue)
                storage = FileSystemStorage();
                storage.set(filename, self, tmpfile);
                contenttype=storage.get(filename,self).getContentType()
            elif HAS_BLOB:
                if isinstance(submittedValue, FileUpload) or type(submittedValue) == file:
                    submittedValue.seek(0)
                    contenttype = guessMimetype(submittedValue, filename)
                    submittedValue = submittedValue.read()
                try:
                    blob = BlobWrapper(contenttype)
                except:
                    # BEFORE PLONE 4.0.1
                    blob = BlobWrapper()
                file_obj = blob.getBlob().open('w')
                file_obj.write(submittedValue)
                file_obj.close()
                blob.setFilename(filename)
                blob.setContentType(contenttype)
                self._setObject(filename, blob)
            else:
                self.manage_addFile(filename, submittedValue)
                contenttype=self[filename].getContentType()
            return (filename, contenttype)
        else:
            return (None, "")

    security.declareProtected(EDIT_PERMISSION, 'deletefile')
    def deletefile(self, filename):
        """
        """
        if(self.getParentDatabase().getStorageAttachments()):
            storage = FileSystemStorage();
            storage.unset(filename, self);
        else:
            if filename in self.objectIds():
                self.manage_delObjects(filename)

    security.declarePublic('isNewDocument')
    def isNewDocument(self):
        """
        """
        req = getattr(self, 'REQUEST', None)
        if req and '/createDocument' in req['ACTUAL_URL']:
            return True
        return False

    security.declarePublic('isDocument')
    def isDocument(self):
        """
        """
        return True

    security.declarePublic('isEditMode')
    def isEditMode(self):
        """
        """
        # if REQUEST exists, test the current command
        if hasattr(self, 'REQUEST'):
            command=self.REQUEST.URL.split('/')[-1].lower()
            return command in ['editdocument', 'edit']
        else:
            return False

    def generateNewId(self):
        """ compute the id using the Document id formula
        (overwrite Archetypes generateNewId, and provides same behaviour, 
        but use Document id formula instead of title)
        """
        form = self.getForm()
        if not form:
            return None

        if not form.getDocumentId():
            return None

        result = None
        try:
            result = self.runFormulaScript("form_"+form.id+"_docid", self, form.DocumentId)
        except PlominoScriptException, e:
            e.reportError('Document id formula failed')

        if not result:
            return None

        # Don't do anything without the plone.i18n package
        if not URL_NORMALIZER:
            return None

        if not isinstance(result, unicode):
            charset = self.getCharset()
            result = unicode(result, charset)

        request = getattr(self, 'REQUEST', None)
        if request is not None:
            return IUserPreferredURLNormalizer(request).normalize(result)

        return queryUtility(IURLNormalizer).normalize(result)
    
registerType(PlominoDocument, PROJECTNAME)
# end of class PlominoDocument

##code-section module-footer #fill in your manual code here
class TemporaryDocument(PlominoDocument):

    security = ClassSecurityInfo()

    def __init__(self, parent, form, REQUEST, real_doc=None):
        self._parent=parent
        if real_doc is not None:
            self.items=real_doc.items.copy()
            self.real_id=real_doc.id
        else:
            self.items={}
            self.real_id="TEMPDOC"
        self.setItem('Form', form.getFormName())
        form.readInputs(self, REQUEST)
        self._REQUEST=REQUEST

    security.declarePublic('getParentDatabase')
    def getParentDatabase(self):
        """ Temporary docs are able to acquire from the db
        """
        return self._parent
        
    security.declarePublic('isEditMode')
    def isEditMode(self):
        """
        """
        return True

    security.declarePublic('isNewDocument')
    def isNewDocument(self):
        """
        """
        if self.real_id=="TEMPDOC":
            return True
        else:
            return False

    security.declarePublic('REQUEST')
    @property
    def REQUEST(self):
        """
        """
        return self._REQUEST
    
    security.declarePublic('id')
    @property
    def id(self):
        """
        """
        return self.real_id

##/code-section module-footer






# -*- coding: utf-8 -*-
#
# File: PlominoAccessControl.py
#
# Copyright (c) 2008 by ['Eric BREHAULT']
#
# Zope Public License (ZPL)
#

__author__ = """Eric BREHAULT <eric.brehault@makina-corpus.com>"""
__docformat__ = 'plaintext'

from Products.ZCatalog.ZCatalog import Catalog
from Missing import MV
from Products.CMFPlomino.PlominoUtils import asUnicode

try:
    from DocumentTemplate.cDocumentTemplate import safe_callable
except ImportError:
    # Fallback to python implementation to avoid dependancy on DocumentTemplate
    def safe_callable(ob):
        # Works with ExtensionClasses and Acquisition.
        if hasattr(ob, '__class__'):
            return hasattr(ob, '__call__') or isinstance(ob, types.ClassType)
        else:
            return callable(ob)

class PlominoCatalog(Catalog):
    """Plomino catalog (just overloads recordify method)
    """

    def recordify(self, object):
        """ turns an object into a record tuple """
        record = []
        # the unique id is always the first element
        for x in self.names:
            if x.startswith("PlominoViewColumn_"):
                param = x.split('_')
                viewname = param[1]
                columnname = param[2]
                if not object.isSelectedInView(viewname):
                    v = None
                else:
                    v = asUnicode(object.computeColumnValue(viewname, columnname))
                record.append(v)
            else:
                attr=getattr(object, x, MV)
                if attr is not MV and safe_callable(attr):
                    attr=attr()
                if type(attr) == type(''):
                    attr = attr.decode('utf-8')
                elif type(attr) == type([]) or type(attr) == type(()):
                    new_value = []
                    for v in attr:
                        if type(v) == type(''):
                            v = v.decode('utf-8')
                        new_value.append(v)
                    attr = new_value
                record.append(attr)
        return tuple(record)
